import os
import datetime

from gites.subpackage.datastore_json_handler import DatastoreJSONHandler
from gites.subpackage._subprocess_handler import run



# A concise interface function to other module:  
def bulkpush(list_of_repo):
    return GitPushManager.listpush(list_of_repo)

# Complete structure: 
class GitPushManager:
    def __init__(self):
        # granting all informations: 
        self.datastore_json_handler = DatastoreJSONHandler()
        self.root_folder = self.datastore_json_handler.root_dir
        
        all_items = os.listdir(self.root_folder)
        self.repo_list  = [os.path.join(".", item) for item in all_items if os.path.isdir(os.path.join(self.root_folder, item)) and not item.startswith('.')]
        
        # for statistics and summary: 
        self.success_repo = []
        self.failed_repo = []
        self.no_effect_repo = []
        self.repo_that_remote_has_new_update = []
        
        # go to root folder before list pushing. 
        os.chdir(self.root_folder)


    def lpush(self):

        print(f"Totally there are {len(self.repo_list)} repos to work on. They are: {self.repo_list}")
        os.chdir(self.root_folder)

        for repo in self.repo_list:
            """
            chain of commands for each repo 
            (if failed, stop and skip to the next repo):

            git fetch
            git checkout main
            git status
            git add --all
            git commit -m {tag}
            git push origin main
            """
            
            print("+" * 72)
            print(f"Current working directory: {os.getcwd()}")
            print(f"Working on repository: {repo}")
            
            # Git fetch
            return_code, stdout = run(["git", "fetch"], loc = repo)
            if return_code != 0:
                self.no_effect_repo.append(repo)
                self.repo_that_remote_has_new_update.append(repo)
                message = f"""
                Everytime before pushing changes for any repo, gites would fetch, and new updates on the remote server might be found. 
                This package will collect all repos that there are updates in the remote server, as a list. 
                
                Step 1: 
                You might use 'gites lpull' to pull all repos in your root folders, if any particular repos only has been changed in the remote server,
                but no changes in the local folder. That command will gives you a summary of which folders are pulled, which folders has version conclict.

                Step 2:
                After step 1, if there are still repos has conflict, you might use `git pull` to each repos that has been indicated as `version conflict`, and fix them one by one. 
                """
                print(message)
                # Continue: terminate the process for this element, proceed next element in the for-loop
                continue 

            # Git checkout
            return_code, _ = run(["git", "checkout", "main"], loc = repo)
            if return_code != 0:
                self.no_effect_repo.append(repo)
                print("Git checkout failed")
                continue 

            # Git status
            return_code, stdout = run(["git", "status"], loc=repo)
            # 3 cases in 'git status': Case 1: there is something new (no need to stop). Case 2 and 3: stop the `listpush` for that repo.
            # Case 3: Other case 
            if return_code != 0:
                self.failed_repo.append(repo)
                print("Git status command failed.")
                continue # type: ignore
            # Case 2: 'Your branch is up to date'
            if "nothing to commit, working tree clean" in stdout:
                self.no_effect_repo.append(repo)
                print("This repo is new. Proceed the next repo.")
                continue # type: ignore

            # Git add all
            return_code, _ = run(["git", "add", "--all"], loc=repo)
            if return_code != 0:
                self.failed_repo.append(repo)
                print("Failed. Code: a1sd32")
                continue # type: ignore

            # Git commit
            tag_message="Automated add-commit-push"
            timetag_for_commit = datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")
            commit_command = ["git", "commit", "-m", f"{tag_message}. Datetime tag: {timetag_for_commit}"]
            return_code, stdout = run(commit_command, loc=repo)
            if return_code != 0:
                self.failed_repo.append(repo)
                print("Git status command failed.")
                continue # type: ignore
            # Case 2: Any other case
            if "Your branch is up to date" in stdout:
                print("No changes to commit.")
                self.no_effect_repo.append(repo)
                print("No effect to the repo. Code: 1sa3a")
                continue
            if return_code != 0:
                self.failed_repo.append(repo)
                print("Failed. Code: 1zsxer2")
                continue # type: ignore
            # Case 3 (Trivial): The staging area is empty (But it is most probably blocked due to the previous `git add` control flow)
            if "nothing to commit" in stdout:
                print("No changes to commit.")
                self.no_effect_repo.append(repo)
                print("No effect to the repo. Code: t5gs2")
                continue # type: ignore
            print(f"Commit successful. Output:\n{stdout}")

            # Git push origin main - 
            push_command = ["git", "push", "origin", "main"]
            _, push_output = run(push_command, loc=repo)
            # Case 3: if there is any error returning code
            if return_code != 0:
                self.failed_repo.append(repo)
                print("Git status command failed.")
                continue # type: ignore
            # Case 2: A summary message returned by git push, indicating a success push
            if "Total" in push_output or not push_output:
                self.success_repo.append(repo)
                # Future development: Scrape the terminal output and collect those data to summary variable.
                print("Push completed")
            # Case 1: Any other case
            else:
                print(f"Push failed. Output:\n{push_output}")
                self.failed_repo.append(repo)
            print("=" * 72)

            
        print("|" * 72)
        print("Add-commit-push completed for all repos in the root folder. Here is the work summary: ")
        print(f"Successful repos (Totally {len(self.success_repo)}): {self.success_repo}")
        print(f"Failed repos (Totally {len(self.failed_repo)}): {self.failed_repo}")
        print(f"No effect repos (Totally {len(self.no_effect_repo)}): {self.no_effect_repo}")
        print(f"Repos that has new updates in the remote server (Totally {len(self.repo_that_remote_has_new_update)}): {self.repo_that_remote_has_new_update}")


# Testing unit: 
if __name__ == "__main__":
    GitPushManager().lpush()




"""
This is a private module that encapsulate the communication process from the JSONUpdater. 
This module should not contains main functions that the package would use. 
"""

import os
import json
from .config_json_handler import ConfigJSONHandler

### Complete structure: 
class DatastoreJSONHandler:
    def __init__(self):
        self.config_json_handler = ConfigJSONHandler()
        # the following are the data from the datastore json: 
        self.gites_datastore_json_location = self.config_json_handler.gites_datastore_json_location # this is for naming convention.
        self.data = self.load_datastore_json()
        self.list_of_repo_details = self.data["repositories"] # this return a list of dictionary, containing 'name' and 'remote_url' key for each repo.
        self.root_dir = self.get_root_directory()

    def get_root_directory(self):
        # Determine the appropriate root directory based on the operating system
        if os.name == 'posix':  # Unix/Linux/Mac OS
            return self.data["root_directories"]["linux"]
        elif os.name == 'nt':  # Windows
            return self.data["root_directories"]["windows"]
        else:
            raise OSError("Unsupported operating system")

    def load_datastore_json(self):
        existing_data = {}

        # case 1: that location has exactly-matched file
        if os.path.exists(self.gites_datastore_json_location):
            with open(self.gites_datastore_json_location, "r") as json_file:
                try:
                    existing_data = json.load(json_file)
                except json.JSONDecodeError as e:
                    print(f"Error loading JSON: {e}")
        else:
            raise SystemExit(f"Terminating due to the datastore json file is not existing in: {self.gites_datastore_json_location}")
        # case 2a: after opening the file, the outermost structure is not a json
        if not isinstance(existing_data, dict):
            print("Please check the file. The whole file is not detected as a json file.")
            raise SystemExit("Terminating due to unexpected error")
        # case 2b: Check each item in the "repositories" data to verify it is well-structured. 
        for repo_detail in existing_data["repositories"]:
            if not isinstance(repo_detail, dict):
                print("Please check the file. The repository list of the file is not detected as a json file.")
                raise SystemExit("Terminating due to unexpected error")
        # case 3: after knowing it is a json, but the json has empty repo entries
        if "repositories" not in existing_data:
            print("it is an empty json. creating structure for it...")
            existing_data["repositories"] = []
        # case 4: if "root_directory" is not specified in the datastore json.
        if "root_directories" not in existing_data:
            raise SystemExit(f"Terminating due to the root directory of your git repository folders are not properly defined in your datastore json: {self.gites_datastore_json_location}")
        # finally return data
        return existing_data

# Testing unit: 
if __name__ == "__main__":
    # this module need an existing json, or existing data, to perform action. 
    # default_json_path = DatastoreLocationChecker().load_setup_json_and_get_datastore_json_address()
    # json_existing_data = _JSONHandler(default_json_path)
    # root_dir = json_existing_data.data.get("root_directory")
    # this is just a default location. 
    pass


import os

from ._subprocess_handler import run
from ._timing import timing
from gites.subpackage.datastore_json_handler import DatastoreJSONHandler



# Complete structure: 
class RepoCloner:
    def __init__(self):
        self.datastore_json_handler = DatastoreJSONHandler()
        self.root_folder = self.datastore_json_handler.root_dir
        self.list_of_repo_details = self.datastore_json_handler.list_of_repo_details

        """
        # old settings
        self.config_file_path = config_file_path
        self.root_directory = ""
        self.repositories = []
        """

        self.success = []
        self.failed = []
        self.no_effect = []

    @timing
    def lclone(self):
        # self.load_config()
        self.clone_repositories()
        self.print_summary()

    """
    def load_config(self):
        with open(self.config_file_path, "r") as config_file:
            config = json.load(config_file)
        self.root_directory = config["root_directory"]
        self.repositories = config["repositories"]
    """

    def clone_repo(self, remote_url, local_path):
        
        # check whether that folder already exists AND it is not empty.
        # It won't bother any existing folder. 
        if os.path.exists(local_path) and any(os.listdir(local_path)):
            print(f"Repository '{local_path}' already exists. Skipping cloning.")
            return 2
        # Don't need to create folder, because git clone would take care of it.
        
        # execute git command
        # go to the root folder
        os.chdir(self.root_folder)
        command = ["git", "clone", remote_url]
        return_code, _ = run(command, loc = self.root_folder)
        
        # to verify the clone, by checking whether the `.git` exist.
        git_folder_path = os.path.join(local_path, ".git")
        if not os.path.exists(git_folder_path):
            print(f"Error: '.git' folder was not generated for repository '{local_path}'. Aborting.")
            # could put a strong SystemExit here
        
        # return status code 
        return return_code

    def clone_repositories(self):
        
        # go to the root folder
        os.chdir(self.root_folder)
        
        for repo_info in self.list_of_repo_details:
            repo_name = repo_info["name"]
            repo_url = repo_info["remote_url"]
            local_path = os.path.join(self.root_folder, repo_name)
            status_message = self.clone_repo(repo_url, local_path)
            if status_message == 2:
                self.no_effect.append(repo_name)
                continue
            elif status_message !=0:
                self.failed.append(repo_name)
                continue
            elif status_message == 0:
                self.success.append(repo_name)
                print(f"Repository '{repo_name}' cloned successfully.")

    def print_summary(self):
        print("Repositories cloned successfully:", self.success)
        print("Repositories failed to clone:", self.failed)
        print("Repositories already existed:", self.no_effect)

if __name__ == "__main__":
    cloner = RepoCloner()
    cloner.lclone()


"""
The followings are legacy functions. No application for now. 
"""

import os
import subprocess
import shutil



# This should be used in more than 1 module.... Check it. 
def check_and_copy_pre_commit_hook(repo_location):
    hooks_dir = os.path.join(repo_location, ".git", "hooks")
    pre_commit_hook_path = os.path.join(hooks_dir, "pre-commit")
    pre_made_hook_path = os.path.join(os.path.dirname(__file__), "My hooks", "pre-commit")  # Replace this with the actual path of your pre-made hook.

    if not os.path.exists(hooks_dir):
        os.makedirs(hooks_dir)

    if not os.path.exists(pre_commit_hook_path):
        print("Copying pre-commit hook...")
        shutil.copy(pre_made_hook_path, pre_commit_hook_path)
        os.chmod(pre_commit_hook_path, 0o755)
        print("Pre-commit hook copied.")
    else:
        print("Pre-commit hook already exists.")

def get_files_bigger_than_100mb(folder_path):
    file_list = []

    for root, _, files in os.walk(folder_path):
        for file_name in files:
            file_path = os.path.join(root, file_name)
            file_size_mb = os.path.getsize(file_path) / (1024 * 1024)  # Convert bytes to megabytes

            if file_size_mb > 100:
                file_list.append(os.path.relpath(file_path, start=folder_path))

    return file_list

def clean_undone_commit():
    """When crashes (eg: some large files included in the commit), clean undone commit by this."""
    subprocess.run(f"git reset --soft HEAD~", shell=True)


def check_ignored_path():
    """Before this, make sure .gitignore file is located in the root directory of the repository."""
    subprocess.run(f"git status --ignored", shell=True)

def print_folder_structure(folder_path, level=0, num_files_to_print=-1, max_depth=float('inf'), current_depth=0, ignore_dot_folders=True):
    # Default depth: infinite
    if current_depth > max_depth:
        return

    # Get the files and directories in the current directory
    items = os.listdir(folder_path)

    # Loop through the items in the directory
    for i, item in enumerate(sorted(items)):
        if ignore_dot_folders and item.startswith('.'):
            continue
        
        item_path = os.path.join(folder_path, item)

        if os.path.isfile(item_path):
            if num_files_to_print < 0 or i < num_files_to_print:
                print(" " * level * 4 + "- " + item)
            elif i == num_files_to_print:
                print(" " * level * 4 + "- ...")
        else:
            print(" " * level * 4 + "- " + item)

        # Recursively print the contents of subfolders
        if os.path.isdir(item_path):
            print_folder_structure(item_path, level + 1, num_files_to_print, max_depth, current_depth + 1, ignore_dot_folders)


if __name__ == "__main__":
    print_folder_structure(os.getcwd(), max_depth=2,ignore_dot_folders=True)


"""
This cli.py is just an interface that collect and manage all the actions,
which help centralize the logic for handling variaous commands and make it easier to maintain 
and expand in the future. 

There are two parts need to be edited to make CLI works. 
1. cli.py: it is an adapter between CLI listener in `setup.py` and the functionalities from subpackages.
argparse module parse 'subcommands' from command-line arguments. 


2. setup.py: the `entry_points` attribute setup the listeners and the related keywords.
'gites' considered as entry-point and the first keyword of the command to invoke the package.
"""


from .subpackage.config_json_handler import ConfigJSONHandler
from .subpackage.git_push_manager import GitPushManager
from .subpackage.repo_cloner import RepoCloner

import argparse

def _check_datastore_location():
    """This function would be optionally used by multiple CLI actions. """
    datastore_json_path = ConfigJSONHandler.check_initial_setup_then_get_datastore_json_address()

def cli_lpush():
    GitPushManager().lpush() 

def cli_lclone():
    RepoCloner().lclone() 

def main():
    parser = argparse.ArgumentParser(description='Command-line interface for gites package')
    # subparsers handle subcommands. 
    subparsers = parser.add_subparsers(dest='command', help='Available commands')

    # Create a subparser for the 'push' command
    push_parser = subparsers.add_parser('lpush', help='Push changes to Git')
    push_parser.set_defaults(func=cli_lpush)

    clone_parser = subparsers.add_parser('lclone', help='Clone a list of repos from Git to local computer')
    clone_parser.set_defaults(func=cli_lclone)

    args = parser.parse_args()

    if hasattr(args, 'func'):
        args.func()

"""
Also check the CLI entry point script at <package_root_folder>/bin/gites.
There should be a shebang in the first line, to be something like: #!/usr/bin/env python
"""

# The content of that file would be like:
"""
#!/home/kin/anaconda3/bin/python
# -*- coding: utf-8 -*-
import re
import sys
from gites.cli import cli_lpush
if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw|\.exe)?$', '', sys.argv[0])
    sys.exit(cli_lpush())
"""


if __name__ == "__main__":
    cli_lpush()




from setuptools import setup, find_packages

package_data = {
    'gites.subpackage': ['package_setup.json'],
}

setup(
    name='gites',
    version='0.4.0',
    description='gites is the home of all developers. It provides bulk clone, bulk push, bulk pull etc. Contact me with my discord ID: pakkin.lau',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Pak Kin LAU',
    author_email='kinlau529@gmail.com',
    url='https://github.com/pakkinlau/gites',
    packages=find_packages(),
    package_data = package_data,
    include_package_data= True,  
    # ["."] or find_package(): Include all packages under the current directory 
    # before: find_packages(),
    install_requires=[],
    entry_points={
        "console_scripts": [
            "gites=gites.cli:main", 
            # gites: the key is the keyword you will use to invoke the package from the command line. 
            # gites.cli: this refer to the module where the entry point function is location. 
            # cli_lpush: this is the name of function within the module that will be executed when you run gites.
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
    ],
)